# standalone file that can be shared without the rest of the package
from loguru import logger  
import pandas as pd
import xarray as xr
import numpy as np


def read_mat_struct_as_dataset(fname, drop_keys=None, index=None, index_is_datenum=False):
    """
    Read a MATLAB struct from a .mat file and return it as an xarray dataset.

    Parameters
    ----------
    fname : str
        Path to the .mat file. All variables in the struct are assumed to have 
        the same dimensions and shape (except for the index columns).
    drop_keys : list, optional [default=None]
        List of keys to drop from the struct. If None is passed [default], then
        no keys are dropped. This can be used when one of the struct fields is
        not the same shape as the others.
    index : str, tuple, optional [default=None]
        Name of the index column. If None is passed [default], then no index is set.
        If a tuple is passed, then the corresponding columns are used as a multiindex.
    index_is_datenum : bool, optional [default=False]
        If True, then the index is converted from MATLAB datenum to a pandas.Timestamp.
    
    Returns
    -------
    ds : xr.Dataset
        Dataset with the struct fields as variables and the corresponding
        data as values. 
    """
    data = read_mat_struct_flat_as_dict(fname)

    if drop_keys is not None:
        for key in drop_keys:
            data.pop(key, None)

    ds = flat_dict_to_xarray(data, index=index, index_is_datenum=index_is_datenum)

    return ds


def flat_dict_to_xarray(data: dict, index=None, index_is_datenum=False) -> xr.Dataset:
    """
    Convert a flat dictionary to an xarray dataset.

    Parameters
    ----------
    data : dict
        Dictionary with the struct fields as keys and the corresponding
        data as values.
    index : str, tuple, optional [default=None]
        Name of the index column. If None is passed [default], then no index is set.
        If a tuple is passed, then the corresponding columns are used as a multiindex.
    index_is_datenum : bool, optional [default=False]
        If True, then the index is converted from MATLAB datenum to a pandas.Timestamp.
    
    Returns
    -------
    ds : xr.Dataset
        Dataset with the struct fields as variables and the corresponding
        data as values. 
    """
    df = pd.DataFrame.from_dict(data)
    if index is not None:
        df = df.set_index(index)
    ds = df.to_xarray()

    if index_is_datenum:
        assert index is not None, "index must be set if index_is_matlab_datenum is True"
        assert isinstance(index, str), "index must be a single index (dtype string) if index_is_matlab_datenum is True"
        ds = ds.assign_coords(**{index: matlab2datetime(ds[index].values)})

    return ds


def read_mat_struct_flat_as_dict(fname: str, key=None) -> dict:
    """
    Read a MATLAB struct from a .mat file and return it as a dictionary.
    
    Assumes that the struct is flat, i.e. it does not contain any nested
    structs.

    Parameters
    ----------
    fname : str
        Path to the .mat file
    key : str, optional
        The name of the matlab key in the .mat file. If None is passed [default],
        then the first key that does not start with an underscore is used.
        If a string is passed, then the corresponding key is used.
    
    Returns
    -------
    data : dict
        Dictionary with the struct fields as keys and the corresponding
        data as values.
    """
    from scipy.io import loadmat

    raw = loadmat(fname)
    
    keys = [k for k in raw.keys() if not k.startswith('_')]

    if key is None:
        logger.log(5, f"No key specified. Using first key that does not start with an underscore: {keys[0]}")
        key = keys[0]
    elif key not in keys:
        raise ValueError(f"Key '{key}' not found in .mat file. Available keys are: {keys}")
    
    named_array = unnest_matlab_struct_named_array(raw[key])
    data = {k: named_array[k].squeeze() for k in named_array.dtype.names}

    return data
    

def unnest_matlab_struct_named_array(arr: np.ndarray) -> dict:
    """
    Unnest a numpy structured array that was read from a MATLAB .mat file.

    Parameters
    ----------
    arr : np.ndarray
        Structured array read from a .mat file
    
    Returns
    -------
    arr : np.ndarray
        Unnested named array where arr.dtype.names has a length > 1 
        and arr[i] corresponds with arr.dtype.names[i]. Note that 
    
    Note
    ----
    This function works with two of my examples, but may be a bit buggy. 
    It is not well tested and may not work in all cases.
    """
    def is_ndarray_or_void(x):
        return isinstance(x, np.ndarray) or isinstance(x, np.void)
    
    while is_ndarray_or_void(arr) and arr.size == 1:
        if (  # stop if a void array with multiple fields
            isinstance(arr, np.void) 
            and (arr.dtype.names is not None) 
            and (len(arr.dtype.names) > 1)):
            return arr
        # otherwise continue
        prev = arr
        arr = prev[0]

    return prev


def datetime2matlab(time: xr.DataArray, reference_datestr:str="1970-01-01") -> np.ndarray:
    """
    Converts the time dimension of a xarray dataset to matlab datenum format

    Parameters
    ----------
    time_hrs : xr.DataArray
        Time from dataset, but only supports hour resolution and lower (days, months, etc)
    reference_datestr : str
        Reference date string in format 'YYYY-MM-DD'. In many cases this is 1970-01-01

    Returns
    -------
    np.ndarray
        Array of matlab datenum values
    """

    def get_matlab_datenum_offset(reference_datestr):
        """
        Returns the matlab datenum offset for a given reference date string
        """
        
        # this is hard coded in matlab, which uses 0000-01-01 as the reference date
        # but this isn't a valid date in pandas, so we use -0001-12-31 instead
        matlab_t0 = pd.Timestamp('-0001-12-31')  
        reference_date = pd.Timestamp(reference_datestr)
        offset_days = (matlab_t0 - reference_date).days
        
        return offset_days
        
    hours_since_ref = time.values.astype('datetime64[h]').astype(float)
    days_since_ref = hours_since_ref / 24

    matlab_offset = get_matlab_datenum_offset(reference_datestr)
    matlab_datenum = days_since_ref - matlab_offset

    return matlab_datenum


def matlab2datetime(matlab_datenum):
    """
    Convert a MATLAB datenum to a pandas.Timestamp

    Parameters
    ----------
    matlab_datenum : float
        MATLAB datenum
    
    Returns
    -------
    pd.Timestamp
        Timestamp object
    """

    matlab_epoch = 719529
    timestamps = pd.to_datetime(matlab_datenum - matlab_epoch, unit='D')

    return timestamps


# standalone file that can be shared without the rest of the package
import pathlib
import pandas as pd
import numpy as np
from loguru import logger
from munch import Munch


class CryoGridConfigExcel:
    """
    A class to read CryoGrid Excel configuration files and extract file paths
    and maybe in the future do some checks etc
    """
    def __init__(self, fname_xls: str, checks=True):
        """
        fname_xls: str
            Path to the CryoGrid Excel configuration file
        checks: bool
            If True, check if the forcing file name matches the given forcing years
        """
        self.fname = pathlib.Path(fname_xls).resolve()
        self.root = self._get_root_path()
        self._df = self._load_xls(fname_xls)
        logger.success(f"Loaded CryoGrid Excel configuration file: {self.fname}")

        if checks:
            self.check_forcing_fname_times()

        self.fname = Munch()
        self.fname.dem = self.get_dem_path()
        self.fname.coords = self.get_coord_path()
        self.fname.era5 = self.get_forcing_path()
        self.fname.datasets = self.get_dataset_paths()
        
        self.time = self.get_start_end_times()

        logger.info(f"Start and end times: {self.time.time_start:%Y-%m-%d} - {self.time.time_end:%Y-%m-%d}")

    def _get_root_path(self):
        path = self.fname.parent
        while True:
            flist = path.glob('run_cryogrid.m')
            if len(list(flist)) > 0:
                self.root = path
                logger.info(f"Found root path: {path}")
                return self.root
            elif str(path) == '/':
                logger.warning("Could not find root path. Set to current directory. You can change this manually with excel_config.root = pathlib.Path('/path/to/root')")
                return pathlib.Path('.')
            else:
                path = path.parent

    def get_start_end_times(self):
        times = self.get_class('set_start_end_time').T.filter(regex='time')
        times = times.map(lambda x: pd.Timestamp(year=int(x[0]), month=int(x[1]), day=int(x[2])))

        start = times.start_time.min()
        end = times.end_time.max()
        times = pd.Series([start, end], index=['time_start', 'time_end'])

        return times

    def get_coord_path(self):
        fname = self.get_class_filepath('COORDINATES_FROM_FILE', fname_key='file_name', index=1)
        return fname

    def get_dataset_paths(self):
        paths = self.get_class_filepath('READ_DATASET', fname_key='filename').to_frame(name='filepath').T
        
        datasets = self.get_class('READ_DATASET')
        variable = datasets.T.variable_name
        paths.loc['variable'] = variable

        paths = Munch(**paths.T.set_index('variable').filepath.to_dict())
        
        return paths
    
    def get_dem_path(self):
        fname = self.get_class_filepath('DEM', folder_key='folder', fname_key='filename', index=1)
        return fname
    
    def get_forcing_path(self, class_name='read_mat_ERA'):
        fname = self.get_class_filepath(class_name, folder_key='path', fname_key='filename', index=1)
        return fname
    
    def check_forcing_fname_times(self):
        """
        a quick check to see if the file name matches the given forcing years
        """
        import re

        fname = self.get_forcing_path()
        times = self.get_start_end_times().dt.year.astype(str).values.tolist()

        fname_years = re.findall(r'[-_]([12][1089][0-9][0-9])', fname.stem)
        
        assert times == fname_years, f"File name years do not match the forcing years: forcing {times} != fname {fname_years}"
    
    def _load_xls(self, fname_xls: str) -> pd.DataFrame:
        import string

        alph = list(string.ascii_uppercase)
        alphabet_extra = alph + [a+b for a in alph for b in alph]

        df = pd.read_excel(fname_xls, header=None, dtype=str)
        df.columns = [c for c in alphabet_extra[:df.columns.size]]
        df.index = df.index + 1

        return df

    def _get_unique_key(self, key: str, col_value='B'):
        df  = self._df
        idx = df.A == key
        value = df.loc[idx, col_value].values
        if len(value) == 0:
            return None
        elif len(value) > 1:
            raise ValueError(f"Multiple values found for key: {key}")
        else:
            return value[0]
        
    def get_class_filepath(self, key, folder_key='folder', fname_key='file', index=None):

        df = self.get_class(key)

        keys = df.index.values
        folder_key = keys[[folder_key in k for k in keys]]
        fname_key = keys[[fname_key in k for k in keys]]

        assert len(folder_key) == 1, f"Multiple folder keys found: {folder_key}"
        assert len(fname_key) == 1, f"Multiple fname keys found: {fname_key}"

        names = df.loc[[folder_key[0], fname_key[0]]]
        names = names.apply(lambda ser: self.root / ser.iloc[0] / ser.iloc[1])

        if index is None:
            return names
        elif isinstance(index, int):
            return names.loc[f"{key}_{index}"]
        else:
            raise TypeError(f"index must be None or int, not {type(index)}")
        
    def get_class(self, class_name: str):
        df = self._df
        i0s = df.A == class_name
        i0s = i0s[i0s].index.values

        blocks = [self._find_class_block(i0) for i0 in i0s]
        try:
            df = pd.concat(blocks, axis=1)
        except:
            df = blocks
        return df

    def _find_class_block(self, class_idx0: int):
        df = self._df

        class_name = df.A.loc[class_idx0]
        msg = f"Given class_idx0 ({class_name}) is not a class. Must have 'index' adjacent or on cell up and right."
        is_index = df.B.loc[class_idx0 - 1: class_idx0].str.contains('index')
        assert is_index.any(), msg

        index_idx = is_index.idxmax()
        class_idx0 = index_idx

        class_idx1 = df.A.loc[class_idx0:] == 'CLASS_END'
        # get first True occurrence
        class_idx1 = class_idx1.idxmax()
        class_block = df.loc[class_idx0:class_idx1]
        class_block = self._process_class_block(class_block)

        return class_block
    
    def _process_class_block(self, df: pd.DataFrame)->pd.DataFrame:
        """hacky way to process the class block"""
        # drop CLASS_END row
        df = df[df.A != 'CLASS_END']

        # if any cell starts with '>', it is a comment
        df = df.map(lambda x: x if not str(x).startswith('>') else np.nan)

        # drop rows and columns that are all NaN
        df = df.dropna(axis=1, how='all').dropna(axis=0, how='all')
        df = df.astype(str)

        # H_LIST and V_MATRIX are special cases
        # return df
        contains_matrix = df.map(lambda x: 'MATRIX' in x).values
        contains_end = df.map(lambda x: 'END' in x).values
        if contains_matrix.any():
            ends = np.where(contains_end)
            r0, c0 = [a[0] for a in np.where(contains_matrix)]

            assert c0 == 1, "Matrix must be in second column"
            assert len(ends) == 2, "Only two ENDs are allowed"
            assert r0 == ends[0][0]
            assert c0 == ends[1][1]

            r1 = ends[0][1]
            c1 = ends[1][0]
            
            arr = df.iloc[r0:r1, c0:c1].values

            matrix = pd.DataFrame(arr[1:, 1:], index=arr[1:, 0], columns=arr[0, 1:])
            matrix.index.name = matrix.columns.name = df.iloc[r0, 0]
            df = df.drop(index=df.index[r0:r1+1])
            df.loc[r0, 'A'] = matrix.index.name
            df.loc[r0, 'B'] = matrix.to_dict(),
        
        for i, row in df.iterrows():
            # H_LIST first
            if row.str.contains('H_LIST').any():
                r0 = 2
                r1 = row.str.contains('END').argmax()
                df.loc[i, 'B'] = row.iloc[r0:r1].values.tolist()

        class_category = df.A.iloc[0]
        class_type = df.A.iloc[1]
        class_index = df.B.iloc[1]
        col_name = f'{class_type}_{class_index}'

        df = (
            df
            .iloc[2:, :2]
            .rename(columns=dict(B=col_name))
            .set_index('A'))
        df.index.name = class_category

        return df


# depends on matlab_helpers.py
import xarray as xr


def _era5_mat_dict_to_xarray(era5_dict: dict)->xr.Dataset:
    """
    Convert a dictionary with the ERA5 forcing variables to a xarray Dataset

    Parameters
    ----------
    era5_dict : dict
        Dictionary with the ERA5 forcing variables. Must contain keys: 
            lon, lat, p, t, dims, 
            u10, v10, u, v, Td2, T2, T, Z, q, P, ps, SW, LW, S_TOA, Zs,
            wind_sf, q_sf, ps_sf, rad_sf, T_sf, P_sf

    Returns
    -------
    xr.Dataset
        Dataset with the ERA5 forcing variables and variables are scaled to the original units
    """
    dat = era5_dict

    out = xr.Dataset(
        coords=dict(
            lon=dat['lon'],
            lat=dat['lat'],
            level=dat['p'],
            time=dat['t']))

    for key in dat:
        if key in ['lon', 'lat', 'p', 't', 'dims']:
            continue
        
        arr = dat[key]
        ndim = len(arr.shape)

        if ndim == 0:
            dims = []
        elif ndim == 2:
            dims = ['lon', 'lat']
        elif ndim == 3:
            dims = ['lon', 'lat', 'time']
        elif ndim == 4:
            dims = ['lon', 'lat', 'level', 'time']
        else:
            raise ValueError(f'Unexpected number of dimensions for `{key}`: {ndim}')
        
        out[key] = xr.DataArray(arr, dims=dims)

    out['u10'] = out['u10'] * dat['wind_sf']
    out['v10'] = out['v10'] * dat['wind_sf']
    out['u'] = out['u'] * dat['wind_sf']
    out['v'] = out['v'] * dat['wind_sf']

    out['Td2'] = out['Td2'] * dat['T_sf']
    out['T2'] = out['T2'] * dat['T_sf']
    out['T'] = out['T'] * dat['T_sf']

    out['SW'] = out['SW'] * dat['rad_sf']
    out['LW'] = out['LW'] * dat['rad_sf']
    out['S_TOA'] = out['S_TOA'] * dat['rad_sf']
    
    out['ps'] = out['ps'] * dat['ps_sf']
    out['P'] = out['P'] * dat['P_sf']
    out['q'] = out['q'] * dat['q_sf']

    out['Zs'] = out['Zs'].astype(float)
    out['Z'] = out['Z'].astype(float)

    out = out.transpose('time', 'level', 'lat', 'lon')

    return out


def read_mat_ear5(filename: str)->xr.Dataset:
    """
    Read the ERA5.mat forcing file for CryoGrid and return a xarray Dataset

    Parameters
    ----------
    filename : str
        Path to the ERA5.mat file

    Returns
    -------
    xr.Dataset
        Dataset with the variables from the ERA5.mat file
    """
    import pathlib
    from .matlab_helpers import read_mat_struct_flat_as_dict

    filename = pathlib.Path(filename).expanduser().absolute().resolve()

    dat = read_mat_struct_flat_as_dict(filename)
    out = _era5_mat_dict_to_xarray(dat)

    out = out.assign_attrs(
        info=(
            'Data read in from CryoGrid ERA5 forcing file. '
            'Data has been scaled to the original units. '
            'Data has been transposed from [lon, lat, level, time] --> [time, level, lat, lon]. '
            'See the ERA5 documentation for more info about the units etc.'),
        source=filename,
    )

    return out


def era5_to_matlab(ds: xr.Dataset)->dict:
    """
    Converrt a merged netCDF file from the Copernicus CDS to 
    a dictionary that matches the expected format of 
    the CryoGrid.POST_PROC.read_mat_ERA class (in MATLAB)

    Parameters
    ----------
    ds : xr.Dataset
        Dataset from the ERA5 Copernicus CDS with variables required for 
        the CryoGrid.POST_PROC.read_mat_ERA class 
        single_levels   = [u10, v10, sp, d2m, t2m, ssrd, strd, tisr, tp, Zs]
        pressure_levels = [t, z, q, u, v]
        Note that Zs in the single levels is a special case since it is only 
        downloaded for a single date at the surface (doesn't change over time)

    Returns
    -------
    dict
        Dictionary with the variables mapped to names that are expected by 
        CryoGrid.POST_PROC.read_mat_ERA
    """
    from .matlab_helpers import datetime2matlab
    import numpy as np

    # transpose to lon x lat x time (original is time x lat x lon)
    ds = ds.transpose('longitude', 'latitude', 'level', 'time')

    era = dict()
    era['dims'] = 'lon x lat (x pressure_levels) x time'
    # while lat and lon have to be [coord x 1]
    era['lat'] = ds['latitude'].values[:, None]
    era['lon'] = ds['longitude'].values[:, None]
    # pressure levels have to be [1 x coord] - only when pressure_levels present
    era['p'] = ds['level'].values[None] * 100 
    # time for some reason has to be [1 x coord]
    era['t'] = datetime2matlab(ds.time)[None] 
    # geopotential height at surface
    era['Zs'] = ds.Zs.values / 9.81  # gravity m/s2

    # single_level variables
    # wind and pressure (no transformations)
    era['u10'] = ds['u10'].values
    era['v10'] = ds['v10'].values
    era['ps'] = ds['sp'].values
    # temperature variables (degK -> degC)
    era['Td2'] = ds['d2m'].values - 273.15
    era['T2'] = ds['t2m'].values - 273.15
    # radiation variables (/sec -> /hour)
    era['SW'] = ds['ssrd'].values / 3600
    era['LW'] = ds['strd'].values / 3600
    era['S_TOA'] = ds['tisr'].values / 3600
    # precipitation (m -> mm)
    era['P'] = ds['tp'].values * 1000

    # pressure levels
    era['T'] = ds['t'].values - 273.15  # K to C
    era['Z'] = ds['z'].values / 9.81  # gravity m/s2
    era['q'] = ds['q'].values
    era['u'] = ds['u'].values
    era['v'] = ds['v'].values

    # scaling factors
    era['wind_sf'] = 1e-2
    era['q_sf'] = 1e-6
    era['ps_sf'] = 1e2
    era['rad_sf'] = 1e-1
    era['T_sf'] = 1e-2
    era['P_sf'] = 1e-2

    # apply scaling factors (done in the original, so we do it here)
    # wind scaling
    era['u']     = (era['u']     / era['wind_sf']).astype(np.int16)
    era['v']     = (era['v']     / era['wind_sf']).astype(np.int16)
    era['u10']   = (era['u10']   / era['wind_sf']).astype(np.int16)
    era['v10']   = (era['v10']   / era['wind_sf']).astype(np.int16)
    # temperature scaling
    era['T']     = (era['T']     / era['T_sf']   ).astype(np.int16)
    era['Td2']   = (era['Td2']   / era['T_sf']   ).astype(np.int16)
    era['T2']    = (era['T2']    / era['T_sf']   ).astype(np.int16)
    # humidity scaling
    era['q']     = (era['q']     / era['q_sf']   ).astype(np.uint16)
    # pressure scaling
    era['ps']    = (era['ps']    / era['ps_sf']  ).astype(np.uint16)
    # radiation scaling
    era['SW']    = (era['SW']    / era['rad_sf'] ).astype(np.uint16)
    era['LW']    = (era['LW']    / era['rad_sf'] ).astype(np.uint16)
    era['S_TOA'] = (era['S_TOA'] / era['rad_sf'] ).astype(np.uint16)
    # precipitation scaling
    era['P']     = (era['P']     / era['P_sf']   ).astype(np.uint16)
    # no scaling for geoportential height
    era['Z']     = era['Z'].astype(np.int16)

    return {'era': era}



# depends on matlab_helpers.py
import xarray as xr
import pandas as pd
import numpy as np

from loguru import logger


def read_OUT_regridded_FCI2_file(fname, deepest_point=None)->xr.Dataset:
    """
    Read a CryoGrid OUT_regridded_FCI2 file and return it as an xarray dataset.

    Parameters
    ----------
    fname : str
        Path to the .mat file
    deepest_point : float, optional
        Represents the deepest depth of the profile. If not provided, 
        then elevation is returned. Negative values represent depths below
        the surface.

    Returns
    -------
    ds : xarray.Dataset
        Dataset with dimensions 'time' and 'level'. The CryoGrid variable
        `depths` is renamed to `elevation`. If deepest_point is provided, then
        `depth` will represent the depth below the surface (negative below
        surface).
    """
    from .matlab_helpers import read_mat_struct_flat_as_dict, matlab2datetime
    
    dat = read_mat_struct_flat_as_dict(fname)
    for key in dat:
        dat[key] = dat[key].squeeze()

    ds = xr.Dataset()
    ds.attrs['filename'] = fname

    times = matlab2datetime(dat.pop('timestamp'))
    elevation = dat.pop('depths')

    for key in dat:
        ds[key] = xr.DataArray(
            data = dat[key].astype('float32'), 
            dims=['level', 'time'], 
            coords={'time': times})
        
    ds['elevation'] = xr.DataArray(elevation, dims=['level'])

    if deepest_point is not None:
        assert deepest_point < 0, "deepest_point must be negative (below surface)"
        
        ds = ds.rename(level='depth')

        # calculate depth step size
        n = elevation.size - 1
        s = (elevation[-1] - elevation[0]) / n
        
        # calculate shallowest point based on step size and n
        shallowest_point = deepest_point - (s * n)
        deepest_point += s / 2  # adding half step for arange
        depth = np.arange(shallowest_point, deepest_point, s)
        ds['depth'] = xr.DataArray(depth, dims=['depth'])
        ds = ds.set_coords('depth')
        ds = ds.transpose('depth', 'time', ...)
    else:  
        ds = ds.transpose('elevation', 'time', ...)

    ds = ds.chunk(dict(time=-1))

    return ds


def read_OUT_regridded_FCI2_parallel(fname_glob, deepest_point, concat_dim='time', **joblib_kwargs):
    """
    Reads multiple files that are put out by the OUT_regridded_FCI2 class

    Parameters
    ----------
    fname_glob: str
        Path of the files that you want to read in. 
        Use same notation as for glob(). Note that it expects
        name to follow the format `some_project_name_GRIDCELL_date.mat`
        where GRIDCELL will be extracted to assign the gridcell dimension. 
        These GRIDCELLs correspond with the index of the data in the 
        flattened array. 
    deepest_point: float
        When setting the configuration for when the data should be 
        saved, the maximum depth is set. Give this number as a
        negative number here.
    concat_dim: str
        The dimension that the data should be concatenated along. 
        Defaults to 'time', but 'gridcell' can also be used if 
        the files are from different gridcells.
    joblib_kwargs: dict
        Uses the joblib library to do parallel reading of the files. 
        Defaults are: n_jobs=-1, backend='threading', verbose=1

    Returns
    -------
    xr.Dataset
        An array with dimensions gridcell, depth, time. 
        Variables depend on how the class was configured, but
        elevation will also be a variable. 
    """
    from glob import glob
    import joblib

    # get the file list
    flist = sorted(glob(fname_glob))
    
    # create the joblib tasks
    func = joblib.delayed(read_OUT_regridded_FCI2_file)
    tasks = [func(f, deepest_point) for f in flist]
    
    # set up the joblib configuration
    joblib_props = dict(n_jobs=-1, backend='threading', verbose=1)
    joblib_props.update(joblib_kwargs)
    worker = joblib.Parallel(**joblib_props)
    output = worker(tasks)  # run the tasks
    
    # combine the coordinates - note filename attr is lost in this step
    ds = xr.combine_nested(output, concat_dim=concat_dim)
    
    # transpose data so that plotting is quick and easy
    ds = ds.transpose('depth', 'time', ...)

    return ds


def read_OUT_regridded_FCI2_clusters(fname_glob, deepest_point, **joblib_kwargs):
    """
    Reads multiple files that are put out by the OUT_regridded_FCI2 class

    Parameters
    ----------
    fname_glob: str
        Path of the files that you want to read in. 
        Use same notation as for glob(). Note that it expects
        name to follow the format `some_project_name_GRIDCELL_date.mat`
        where GRIDCELL will be extracted to assign the gridcell dimension. 
        These GRIDCELLs correspond with the index of the data in the 
        flattened array. 
    deepest_point: float
        When setting the configuration for when the data should be 
        saved, the maximum depth is set. Give this number as a
        negative number here.
    joblib_kwargs: dict
        Uses the joblib library to do parallel reading of the files. 
        Defaults are: n_jobs=-1, backend='threading', verbose=1

    Returns
    -------
    xr.Dataset
        An array with dimensions gridcell, depth, time. 
        Variables depend on how the class was configured, but
        elevation will also be a variable. 
    """
    from glob import glob

    # get the file list
    flist = sorted(glob(fname_glob))
    # extract the gridcell from the file name
    gridcell = [int(f.split('_')[-2]) for f in flist]
    
    ds = read_OUT_regridded_FCI2_parallel(fname_glob, deepest_point, concat_dim='gridcell', **joblib_kwargs)
    ds['gridcell'] = xr.DataArray(gridcell, dims=['gridcell'])

    # transpose data so that plotting is quick and easy
    ds = ds.transpose('gridcell', 'depth', 'time', ...)

    return ds



from .matlab_helpers import (
    read_mat_struct_flat_as_dict,
    read_mat_struct_as_dataset)

from .outputs import (
    read_OUT_regridded_FCI2_file,
    read_OUT_regridded_FCI2_clusters)

from .forcing import era5_to_matlab
from .excel_config import CryoGridConfigExcel